from __future__ import annotations

from enum import Enum
from typing import Callable, Optional, TypeVar, Dict

from pydantic import Field
from pydantic.dataclasses import dataclass

from mr_scraper.api.transforms import get_attribute, get_text

Input = TypeVar("Input")
Value = TypeVar("Value")

TransformFunction = Callable[[Input], Value]
GetUrl = Callable[[Dict[str, any]], str]


class ContentProvider(Enum):
    REQUEST = 1
    PUPPETEER = 2


@dataclass
class QueryConfig:
    selector: str
    transform: Optional[TransformFunction] = lambda x: x
    queries: Optional[Dict[str, QueryConfig]] = None


QueryConfig.__pydantic_model__.update_forward_refs()


@dataclass
class ContentConfig:
    provider: ContentProvider
    delay: int = Field(default=2, description="delay before get content")
    url: GetUrl = lambda x: x.get('url')
    transform: Optional[TransformFunction] = None


@dataclass
class ScraperMessage:
    scraper: str
    type: str
    payload: Dict


@dataclass
class ScraperConfig:
    content: ContentConfig
    queries: Dict[str, QueryConfig]
    callback: Optional[Callable[[any, ScraperMessage], any]] = None


def base_url(base: str) -> GetUrl:
    def wrapper(payload: dict):
        return f'{base}{payload["url"]}'

    return wrapper


def compose_transform(function, fields: Dict[str, any]):
    fn = fields.get('transform', lambda x: x)
    return lambda x: fn(function(x))


def default_query_fields(selector: str, **kwargs):
    fields = dict(**kwargs)
    fields['selector'] = selector
    return fields


def query_config(*attrs: str):
    """
    decorate a function that return a dictionary with the parameters of the QueryConfig
    :param attrs: custom attributes
    :return:
    """

    def decorator(function):
        def wrapper(selector: str, **kwargs) -> QueryConfig:
            fields = default_query_fields(selector, **kwargs)
            args = {'fields': fields}
            for v in attrs:
                args[v] = fields.get(v)
            fields = function(**args)
            return QueryConfig(
                selector=fields.get('selector'),
                transform=fields.get('transform'),
                queries=fields.get('queries')
            )

        return wrapper

    return decorator


@query_config('attribute')
def attribute_query(attribute: str, fields: dict):
    fields['transform'] = compose_transform(lambda x: get_attribute(x, attribute), fields)
    return fields


@query_config()
def text_query(fields: dict):
    fields['transform'] = compose_transform(get_text, fields)
    return fields



import importlib
from typing import Callable, Dict

import requests
from bs4 import BeautifulSoup
from bs4.element import Tag
from webpilot import config, robot

from mr_scraper.api.models import ContentProvider, ScraperConfig, QueryConfig, ScraperMessage

PUPPETEER_CHROME_PATH = 'chromium-browser'

GetContent = Callable[[ContentProvider, str, dict], str]
Dispatch = Callable[[ScraperMessage], any]


def get_content_using_request(url: str, payload: dict, delay: int = 0) -> str:
    return requests.request("GET", url, payload).text


def get_content_using_puppeteer(url: str, payload: dict, delay: int = 0) -> str:
    pilot_config = config.WebPilotConfig(
        headless=True,
        sandboxed=False,  # this is because collab run with root
        remote_port=33333,
        chrome_executable=PUPPETEER_CHROME_PATH)
    with robot.open_chrome(pilot_config) as pilot:
        tab = pilot.new_tab()
        page = pilot.navigate(tab, url, delay)
        return pilot.get_content(page)


__content_providers__ = {
    ContentProvider.REQUEST: get_content_using_request,
    ContentProvider.PUPPETEER: get_content_using_puppeteer
}


def get_content(provider: ContentProvider, url: str, payload: dict, delay: int = 0):
    return __content_providers__[provider](url, payload, delay=delay)


def dispatch(message: ScraperMessage):
    module = importlib.import_module(message.scraper)
    cfg: ScraperConfig = getattr(module, message.type)
    url = cfg.content.url(message.payload)
    content = get_content(cfg.content.provider, url, message.payload, delay=cfg.content.delay)

    result = visit_dom(cfg, content)

    if cfg.callback is not None:
        return cfg.callback(result, message)

    return result


def visit_dom(cfg: ScraperConfig, content: str):
    dom = BeautifulSoup(content, features="html.parser")
    return visit_dom_queries(dom, cfg.queries)


def visit_dom_queries(dom: Tag, queries: Dict[str, QueryConfig]):
    return {key: visit_dom_query(dom, value) for (key, value) in queries.items()}


def visit_dom_query(dom: Tag, value: QueryConfig):
    result = [dom_compound(value.transform(item), value.queries) for item in dom.select(value.selector)]

    if len(result) == 1:
        return result[0]
    return result


def dom_compound(value, queries: Dict[str, QueryConfig]):
    if queries is None:
        return value
    return visit_dom_queries(value, queries)



import pathlib

from setuptools import find_namespace_packages, setup

HERE = pathlib.Path(__file__).parent
README = (HERE / 'README.md').read_text()

setup(
    name='mr_scraper',
    packages=find_namespace_packages(exclude=['build', 'dist', 'example', 'example.scrapers', 'tests']),
    version='0.6.0',
    description='Library to create scrapers in python',
    long_description=README,
    long_description_content_type='text/markdown',
    author='fernandojerez',
    license='Apache',
    requires=['webpilot', 'requests'],
    install_requires=['webpilot', 'requests'],
    test_suite='tests'
)
